﻿using System;
using System.Windows.Forms;
using System.IO;
using System.Drawing;
using System.Collections.Generic;
using STFXmlParser;
using Selections;
using PdfDocument;
using CTNet;

namespace CalcHelper
{
    public partial class Form1 : Form
    {
        public void OpenFile(string zipFile, string vscFile)
        {
            // Should not happen...
            if (string.IsNullOrEmpty(zipFile))
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null or empty zipFile argument:\n\nOpenFile() received a null or empty zipFile argument.\n\nAborting application."))
                {
                    mainMsg.ShowDialog(this);
                }

                Application.Exit();
            }

            // Also should not happen...
            if (string.IsNullOrEmpty(vscFile))
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null or empty vscFile argument:\n\nOpenFile() received a null or empty vscFile argument.\n\nAborting application."))
                {
                    mainMsg.ShowDialog(this);
                }

                Application.Exit();
            }

            docFile = new DocFile(zipFile, vscFile, codeSnippetsFile, fcnSnippetsFile);
            pdfCurDir = Path.GetDirectoryName(zipFile);

            if (docFile.ExtractFiles())
            {
                if (docFile.SetPdfBuffer())
                {
                    if (SetupPdf1(zipFile))
                    {
                        if (splitScreen)
                        {
                            /*
                             * Copy pdf1 over to pdf2, do not fuss with the
                             * .isSelected, .isCalculated, and .isCalculating
                             * fields since we are opening a form.
                            */
                            SetupPdf2(zipFile, false);
                        }
                    }
                }
            }
            else
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.information, string.Format("Calc Helper can not open {0}:\n\nCalc Helper needs BOTH the .pdf and the .xml [boxes] files.\n\nCalc Helper will close.", Path.GetFileName(zipFile))))
                {
                    mainMsg.ShowDialog(this);
                }

                Application.Exit();
            }
        }

        public bool SetupPdf1(string fileName)
        {
            bool result = false;

            pdf1 = new Pdf("pdf1", ref panel1, ref pictureBox1);

            if (docFile.GetXmlData(ref pa1))
            {
                if (pdf1.ScaleRects(ref docFile.hPdf, "OpenFile", ref pa1))
                {
                    if (pdf1.numPages > 0)
                    {
                        docFile.docBoxList = pa1.GetDocBoxList();

                        // Document using new (box24) or old (box2_12) box format?
                        docFile.SetBoxFormat();
                        boxSel.BoxSelSetBoxFormat(docFile.newBoxFormat);

                        if (docFile.ProcessJSFile(false, ref pa1, ref pdf1, ref pa2, ref pdf2))
                        {
                            result = true;

                            docFile.SetCodeNavLists();

                            // Enable appropriate boxed page navigation menu items.
                            EnablePageNavMenuItems(pdf1.id, ref pa1);

                            this.Text = Path.GetFileName(fileName) + " - Calc Helper " + version;
                            docLabel.Text = fileName;

                            pdf1.panel.BackColor = SystemColors.Control;

                            // Will default to Boxed Page navigation mode if enabled.
                            if (boxedPageToolStripMenuItem.Enabled == true)
                            {
                                pdf1.pageNav = Pdf.pageNavMode.boxedpages;
                                pdf1.boxedPageList = pa1.GetBoxedPageList(pdf1.pageNav.ToString());
                                pdf1.boxedPageIndex = 0;
                                pdf1.curPage = pdf1.boxedPageList[0];
                            }
                            else
                            {
                                // Set page navigation to pages:
                                pdf1.pageNav = Pdf.pageNavMode.page;
                                pdf1.curPage = 1;
                            }

                            /*
                             * This sets the text on the first/previous/next/last/page
                             * navigation buttons. For example text will either read 
                             * like 'navigate to next page' or 'navigate to next boxed page'.
                            */
                            ProcessPageNavMode(pdf1.pageNav, pdf1.id);

                            pdf1.ShowPage(ref docFile.hPdf, (int)pdf1.curPage, true, ref pa1);

                            // This handles page navigation menu items/buttons being active/inactive.
                            ProcessPageNum(ref pdf1);

                            /*
                             * Enable non-page navigation and non-Zoom Reset menu items/buttons.
                             * Zoom Reset and page navigation menu buttons/items
                             * are handled by other logic associated with them.
                            */
                            zoomInBtn.Enabled = true;
                            zoomInToolStripMenuItem.Enabled = true;

                            zoomOutBtn.Enabled = true;
                            zoomOutToolStripMenuItem.Enabled = true;

                            fullWidthBtn.Enabled = true;
                            fullWidthToolStripMenuItem.Enabled = true;

                            fullPageBtn.Enabled = true;
                            fullPageToolStripMenuItem.Enabled = true;

                            pageNavigateModeBtn.Enabled = true;

                            // Tools Menu Item:
                            toolsToolStripMenuItem.Enabled = true;

                            if (!splitScreen)
                            {
                                splitHorizBtn.Enabled = true;
                                splitScreenHorizontallyToolStripMenuItem.Enabled = true;

                                splitVertBtn.Enabled = true;
                                splitScreenVerticallyToolStripMenuItem.Enabled = true;
                            }

                            /*
                             * Get the Snippet Extension's code snippet info and function
                             * snippet info. These are displayed on the Tool Belt's Function
                             * Picker and Snippet Picker tab pages.
                             */
                            docFile.GetJSONCodeSnippets();
                            docFile.GetJSONFcnSnippets();

                            ProcessOrphanBoxesButton();
                        }
                    }
                }
            }

            return result;
        }

        public void SetupPdf2(string fileName, bool setBoxSelCal)
        {
            if (pdf1 == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "SetupPdf2(): pdf1 is null."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            pdf2 = new Pdf("pdf2", ref panel2, ref pictureBox2);

            if (docFile.GetXmlData(ref pa2))
            {
                if (pdf2.ScaleRects(ref docFile.hPdf, "OpenFile", ref pa2))
                {
                    if (pdf2.numPages > 0)
                    {
                        // Enable appropriate boxed page navigation menu items.
                        EnablePageNavMenuItems(pdf2.id, ref pa2);

                        docLabel2.Text = fileName;

                        pdf2.panel.BackColor = SystemColors.Control;
                        pdf2.boxedPageList = pdf1.boxedPageList;
                        pdf2.pageNav = pdf1.pageNav;
                        pdf2.boxedPageIndex = pdf1.boxedPageIndex;
                        pdf2.curPage = pdf1.curPage;

                        if (setBoxSelCal)
                        {
                            /*
                             * Set the .isSelected, .isCalculated, and .isCalculating
                             * fields so we show the same info on the split screen.
                            */
                            if (pa1 != null)
                            {
                                if (pa2 != null)
                                {
                                    List<Field> fa1 = null;
                                    List<Field> fa2 = null;
                                    int pa1Length = pa1.Length();

                                    /*
                                     * By definition the pa's and fa's are identical since
                                     * they are both tied to the same document.
                                    */
                                    for (int i = 0; i < pa1Length; i++)
                                    {
                                        pa1.GetFieldArrayRef(i, ref fa1);

                                        if (fa1 != null)
                                        {
                                            pa2.GetFieldArrayRef(i, ref fa2);

                                            for (int j = 0; j < fa1.Count; j++)
                                            {
                                                fa2[j].leftBoxSel = fa1[j].leftBoxSel;
                                                fa2[j].isCalculated = fa1[j].isCalculated;
                                                fa2[j].rightCalcBoxSel = fa1[j].rightCalcBoxSel;
                                                fa2[j].rightCtrlBoxSel = fa1[j].rightCtrlBoxSel;
                                                fa2[j].ctrlType = fa1[j].ctrlType;
                                            }
                                        }
                                    }
                                }
                            }
                        }

                        /*
                         * This sets the text on the first/previous/next/last/page
                         * navigation buttons. For example text will either read 
                         * like 'navigate to next page' or 'navigate to next boxed page'.
                        */
                        ProcessPageNavMode(pdf2.pageNav, pdf2.id);

                        pdf2.ShowPage(ref docFile.hPdf, (int)pdf2.curPage, true, ref pa2);

                        // This handles page navigation menu items/buttons being active/inactive.
                        ProcessPageNum(ref pdf2);

                        /*
                         * Enable non-page navigation and non-Zoom Reset menu items/buttons.
                         * Zoom Reset and page navigation menu buttons/items
                         * are handled by other logic associated with them.
                        */
                        zoomInBtn2.Enabled = true;
                        zoomInToolStripMenuItem2.Enabled = true;

                        zoomOutBtn2.Enabled = true;
                        zoomOutToolStripMenuItem2.Enabled = true;

                        fullWidthBtn2.Enabled = true;
                        fullWidthToolStripMenuItem2.Enabled = true;

                        fullPageBtn2.Enabled = true;
                        fullPageToolStripMenuItem2.Enabled = true;

                        pageNavigateModeBtn2.Enabled = true;
                    }
                }
            }
        }

        public void ProcessPageNum(ref Pdf pdf)
        {
            if (pdf == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null ProcessPdf argument:\n\nProcessPageNum() received a null pdf argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            /*
             * Enable/disable the page navigation toolstrip buttons to reflect
             * current page as it relates to the page navigation mode.
             * 
             * Note that pdf.curPage still holds the current page regardless of
             * the page navigation mode, but we need to look at the pdf.boxedPageList
             * index value when in Boxed Pages mode.
            */

            int curVal, upperVal, lowerVal;

            if (pdf.pageNav == Pdf.pageNavMode.page)
            {
                curVal = (int)pdf.curPage;
                upperVal = (int)pdf.numPages;
                lowerVal = 1;
            }
            else
            {
                curVal = pdf.boxedPageIndex;
                upperVal = (pdf.boxedPageList.Count - 1);
                lowerVal = 0;
            }

            // Each split screen has a plethora of menu buttons:
            switch (pdf.id)
            {
                case ("pdf1"):

                    if (curVal < upperVal)
                    {
                        if (!lastPageBtn.Enabled)
                        {
                            nextPageBtn.Enabled = true;
                            nextPageToolStripMenuItem.Enabled = true;

                            lastPageBtn.Enabled = true;
                            lastPageToolStripMenuItem.Enabled = true;
                        }
                    }
                    else
                    {
                        nextPageBtn.Enabled = false;
                        nextPageToolStripMenuItem.Enabled = false;

                        lastPageBtn.Enabled = false;
                        lastPageToolStripMenuItem.Enabled = false;
                    }

                    if (curVal > lowerVal)
                    {
                        if (!firstPageBtn.Enabled)
                        {
                            prevPageBtn.Enabled = true;
                            previousPageToolStripMenuItem.Enabled = true;

                            firstPageBtn.Enabled = true;
                            firstPageToolStripMenuItem.Enabled = true;
                        }
                    }
                    else
                    {
                        prevPageBtn.Enabled = false;
                        previousPageToolStripMenuItem.Enabled = false;

                        firstPageBtn.Enabled = false;
                        firstPageToolStripMenuItem.Enabled = false;
                    }

                    if (pdf.numPages > 1)
                    {
                        goToPageToolStripMenuItem.Enabled = true;
                    }
                    else
                    {
                        goToPageToolStripMenuItem.Enabled = false;
                    }

                    pageLabel.Text = string.Format("Page {0} of {1}", pdf.curPage, pdf.numPages);

                    break;

                case ("pdf2"):

                    if (curVal < upperVal)
                    {
                        if (!lastPageBtn2.Enabled)
                        {
                            nextPageBtn2.Enabled = true;
                            nextPageToolStripMenuItem2.Enabled = true;

                            lastPageBtn2.Enabled = true;
                            lastPageToolStripMenuItem2.Enabled = true;
                        }
                    }
                    else
                    {
                        nextPageBtn2.Enabled = false;
                        nextPageToolStripMenuItem2.Enabled = false;

                        lastPageBtn2.Enabled = false;
                        lastPageToolStripMenuItem2.Enabled = false;
                    }

                    if (curVal > lowerVal)
                    {
                        if (!firstPageBtn2.Enabled)
                        {
                            prevPageBtn2.Enabled = true;
                            previousPageToolStripMenuItem2.Enabled = true;

                            firstPageBtn2.Enabled = true;
                            firstPageToolStripMenuItem2.Enabled = true;
                        }
                    }
                    else
                    {
                        prevPageBtn2.Enabled = false;
                        previousPageToolStripMenuItem2.Enabled = false;

                        firstPageBtn2.Enabled = false;
                        firstPageToolStripMenuItem2.Enabled = false;
                    }

                    if (pdf.numPages > 1)
                    {
                        goToPageToolStripMenuItem2.Enabled = true;
                    }
                    else
                    {
                        goToPageToolStripMenuItem2.Enabled = false;
                    }

                    pageLabel2.Text = string.Format("Page {0} of {1}", pdf.curPage, pdf.numPages);

                    break;
            }
        }

        public void ProcessPageNavMode(Pdf.pageNavMode pageMode, string pdfId)
        {
            string navToolTipText = "";
            string navText = "";
            string firstText = "";
            string prevText = "";
            string nextText = "";
            string lastText = "";

            switch (pageMode)
            {
                case (Pdf.pageNavMode.page):

                    navToolTipText = "Navigate: To Any Page";
                    navText = "Page";
                    firstText = "First Page";
                    prevText = "Previous Page";
                    nextText = "Next Page";
                    lastText = "Last Page";

                    break;

                case (Pdf.pageNavMode.boxedpages):

                    navToolTipText = "Navigate: To Just Pages With Boxes";
                    navText = "Boxed";
                    firstText = "First Boxed Page";
                    prevText = "Previous Boxed Page";
                    nextText = "Next Boxed Page";
                    lastText = "Last Boxed Page";

                    break;

                case (Pdf.pageNavMode.checkbox):

                    navToolTipText = "Navigate: To Just Pages With Checkboxes";
                    navText = "Checkbox";
                    firstText = "First Checkbox Page";
                    prevText = "Previous Checkbox Page";
                    nextText = "Next Checkbox Page";
                    lastText = "Last Checkbox Page";

                    break;

                case (Pdf.pageNavMode.code39):

                    navToolTipText = "Navigate: To Just Pages With Code39 Barcodes";
                    navText = "Code39";
                    firstText = "First Code39 Page";
                    prevText = "Previous Code39 Page";
                    nextText = "Next Code39 Page";
                    lastText = "Last Code39 Page";

                    break;

                case (Pdf.pageNavMode.int2of5):

                    navToolTipText = "Navigate: To Just Pages With Int2of5 Barcodes";
                    navText = "Int2of5";
                    firstText = "First Int2of5 Page";
                    prevText = "Previous Int2of5 Page";
                    nextText = "Next Int2of5 Page";
                    lastText = "Last Int2of5 Page";

                    break;

                case (Pdf.pageNavMode.pdf417):

                    navToolTipText = "Navigate: To Just Pages With Pdf417 Barcodes";
                    navText = "Pdf417";
                    firstText = "First Pdf417 Page";
                    prevText = "Previous Pdf417 Page";
                    nextText = "Next Pdf417 Page";
                    lastText = "Last Pdf417 Page";

                    break;

                case (Pdf.pageNavMode.qrcode):

                    navToolTipText = "Navigate: To Just Pages With QR Codes";
                    navText = "QR Code";
                    firstText = "First QR Code Page";
                    prevText = "Previous QR Code Page";
                    nextText = "Next QR Code Page";
                    lastText = "Last QR Code Page";

                    break;
            }

            switch (pdfId)
            {
                case ("pdf1"):

                    pageNavigateModeBtn.ToolTipText = navToolTipText;
                    pageNavigateModeBtn.Text = navText;

                    firstPageBtn.ToolTipText = firstText;
                    firstPageToolStripMenuItem.Text = firstText;

                    prevPageBtn.ToolTipText = prevText;
                    previousPageToolStripMenuItem.Text = prevText;

                    nextPageBtn.ToolTipText = nextText;
                    nextPageToolStripMenuItem.Text = nextText;

                    lastPageBtn.ToolTipText = lastText;
                    lastPageToolStripMenuItem.Text = lastText;

                    break;

                case ("pdf2"):

                    pageNavigateModeBtn2.ToolTipText = navToolTipText;
                    pageNavigateModeBtn2.Text = navText;

                    firstPageBtn2.ToolTipText = firstText;
                    firstPageToolStripMenuItem2.Text = firstText;

                    prevPageBtn2.ToolTipText = prevText;
                    previousPageToolStripMenuItem2.Text = prevText;

                    nextPageBtn2.ToolTipText = nextText;
                    nextPageToolStripMenuItem2.Text = nextText;

                    lastPageBtn2.ToolTipText = lastText;
                    lastPageToolStripMenuItem2.Text = lastText;

                    break;
            }
        }

        public void EnableZoomResetBtns(ref Pdf pdf)
        {
            if (pdf == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null ProcessPdf argument:\n\nEnableZoomResetBtns() received a null pdf argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            switch (pdf.id)
            {
                case ("pdf1"):

                    if (!zoomResetBtn.Enabled)
                    {
                        zoomResetBtn.Image = Properties.Resources.ZoomResetOn;
                        zoomResetBtn.Enabled = true;
                        zoomResetToolStripMenuItem.Enabled = true;
                    }

                    break;

                case ("pdf2"):

                    if (!zoomResetBtn2.Enabled)
                    {
                        zoomResetBtn2.Image = Properties.Resources.ZoomResetOn;
                        zoomResetBtn2.Enabled = true;
                        zoomResetToolStripMenuItem2.Enabled = true;
                    }

                    break;
            }
        }

        public void DisableZoomResetBtns(ref Pdf pdf)
        {
            if (pdf == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null ProcessPdf argument:\n\nDisableZoomResetBtns() received a null pdf argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            switch (pdf.id)
            {
                case ("pdf1"):

                    if (zoomResetBtn.Enabled)
                    {
                        zoomResetBtn.Image = Properties.Resources.ZoomResetOff;
                        zoomResetBtn.Enabled = false;
                        zoomResetToolStripMenuItem.Enabled = false;
                    }

                    break;

                case ("pdf2"):

                    if (zoomResetBtn2.Enabled)
                    {
                        zoomResetBtn2.Image = Properties.Resources.ZoomResetOff;
                        zoomResetBtn2.Enabled = false;
                        zoomResetToolStripMenuItem2.Enabled = false;
                    }

                    break;
            }
        }

        public void ProcessZoomIn(ref Pdf pdf, ref PageArray pa)
        {
            if (pdf == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null ProcessPdf argument:\n\nProcessZoomIn() received a null pdf argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            if (pa == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null PageArray argument:\n\nProcessZoomIn() received a null pa argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            PgZoomData zd = null;

            /*
             * First check pdf.initZoomReset. If true, we initialize the page dictionary 
             * with the dpi and page bitmap heights and widths per page. We roll-back to
             * these settings for a zoom reset operation.
            */
            if (pdf.initZoomReset)
            {
                pdf.SetZoomResets(ref pdf, ref pa);
            }

            pdf.ScaleRects(ref docFile.hPdf, "ZoomIn", ref pa);

            if (pdf.curPageZoomDelta)
            {
                pdf.curPageZoomDelta = false;

                pdf.vScrollPos = (int)((double)pdf.panel.VerticalScroll.Value * pdf.zoomInVal);
                pdf.hScrollPos = (int)((double)pdf.panel.HorizontalScroll.Value * pdf.zoomInVal);

                pdf.ShowPage(ref docFile.hPdf, (int)pdf.curPage, true, ref pa);

                pa.GetPgZoomDataRef((int)pdf.curPage - 1, ref zd);

                if (zd == null)
                {
                    using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, string.Format("ProcessZoomIn():\n\npa.GetPgZoomDataRef({0}, ref zd) has returned a null PgZoomData.", pdf.curPage - 1)))
                    {
                        mainMsg.ShowDialog(this);
                    }

                    return;
                }

                /*
                 * We now disable or enable if not enabled already, the zoom reset
                 * buttons for the current page:
                */
                if (zd.zoomCnt == 0)
                {
                    /*
                     * We had prior 'zoom outs' which drove the page's zoom count to
                     * a negative value. This zoom in brought the count up to zero.
                    */
                    DisableZoomResetBtns(ref pdf);
                }
                else
                {
                    EnableZoomResetBtns(ref pdf);
                }

                /*
                 * Trying to set scrollbars in same relative position,
                 * relative to upper left corner of PicPanel.
                 * Need to set these twice for some reason.
                */
                pdf.panel.VerticalScroll.Value = pdf.vScrollPos;
                pdf.panel.VerticalScroll.Value = pdf.vScrollPos;

                pdf.panel.HorizontalScroll.Value = pdf.hScrollPos;
                pdf.panel.HorizontalScroll.Value = pdf.hScrollPos;
            }
        }

        public void ProcessZoomOut(ref Pdf pdf, ref PageArray pa)
        {
            if (pdf == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null ProcessPdf argument:\n\nProcessZoomOut() received a null pdf argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            if (pa == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null PageArray argument:\n\nProcessZoomOut() received a null pa argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            PgZoomData zd = null;

            /*
             * First check pdf.initZoomReset. If true, we initialize the page dictionary 
             * with the dpi and page bitmap heights and widths per page. We roll-back to
             * these settings for a zoom reset operation.
            */
            if (pdf.initZoomReset)
            {
                pdf.SetZoomResets(ref pdf, ref pa);
            }

            pdf.ScaleRects(ref docFile.hPdf, "ZoomOut", ref pa);

            if (pdf.curPageZoomDelta)
            {
                pdf.curPageZoomDelta = false;

                // Current pre-zoom out values.
                pdf.vScrollPos = pdf.panel.VerticalScroll.Value;
                pdf.hScrollPos = pdf.panel.HorizontalScroll.Value;

                pdf.ShowPage(ref docFile.hPdf, (int)pdf.curPage, true, ref pa);

                pa.GetPgZoomDataRef((int)pdf.curPage - 1, ref zd);

                if (zd == null)
                {
                    using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, string.Format("ProcessZoomOut():\n\npa.GetPgZoomDataRef({0}, ref zd) has returned a null PgZoomData.", pdf.curPage - 1)))
                    {
                        mainMsg.ShowDialog(this);
                    }

                    return;
                }

                /*
                 * We now disable or enable if not enabled already, the zoom reset
                 * buttons for the current page:
                */
                if (zd.zoomCnt == 0)
                {
                    /*
                     * We had prior 'zoom ins' which drove the page's zoom count to
                     * a positive value. This zoom out brought the count down to zero.
                    */
                    DisableZoomResetBtns(ref pdf);
                }
                else
                {
                    EnableZoomResetBtns(ref pdf);
                }

                /*
                 * Now zoom-out scrollbar position(s), if still there,
                 * trying to keep them in same relative position(s):
                 * 
                 * Again, we need to set positions twice for some reason.
                */
                if (pdf.bMap.Width > (pdf.panel.Width - 2))
                {
                    pdf.hScrollPos = (int)((double)pdf.hScrollPos * pdf.zoomOutVal);

                    pdf.panel.HorizontalScroll.Value = pdf.hScrollPos;
                    pdf.panel.HorizontalScroll.Value = pdf.hScrollPos;
                }

                if (pdf.bMap.Height > (pdf.panel.Height - 2))
                {
                    pdf.vScrollPos = (int)((double)pdf.vScrollPos * pdf.zoomOutVal);

                    pdf.panel.VerticalScroll.Value = pdf.vScrollPos;
                    pdf.panel.VerticalScroll.Value = pdf.vScrollPos;
                }
            }
        }

        public void ProcessZoomReset(ref Pdf pdf, ref PageArray pa)
        {
            if (pdf == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null ProcessPdf argument:\n\nProcessZoomReset() received a null pdf argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            if (pa == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null PageArray argument:\n\nProcessZoomReset() received a null pa argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            PgZoomData zd = null;
            int zoomCnt;

            // Get the current page's zoom count before we reset.
            pa.GetPgZoomDataRef((int)pdf.curPage - 1, ref zd);

            if (zd == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, string.Format("ProcessZoomReset():\n\npa.GetPgZoomDataRef({0}, ref zd) has returned a null PgZoomData.\n\nAborting Zoom Reset operation.", pdf.curPage - 1)))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            zoomCnt = zd.zoomCnt;

            pdf.ScaleRects(ref docFile.hPdf, "ZoomReset", ref pa);

            /*
             * This will attempt to keep any scrollbars in same relative
             * locations after resetting the zoom.
            */

            // Current pre-zoom out values.
            pdf.vScrollPos = pdf.panel.VerticalScroll.Value;
            pdf.hScrollPos = pdf.panel.HorizontalScroll.Value;

            pdf.ShowPage(ref docFile.hPdf, (int)pdf.curPage, true, ref pa);

            /*
             * Now reset scrollbar position(s), if still there,
             * trying to keep them in same relative position(s):
             * 
             * Again, we need to set positions twice for some reason.
             * We also need to pay attention to the zoomCnt sign,
             * a negative value means we need to zoom in the value
             * while a positive means we zoom out the scroll values.
             * 
             * Note that this is an approximation, especially when
             * this page has been zoomed out to the point of no
             * scroll bars in which case we 'forced' the image to 
             * either a full page or a full width so we lose the
             * connection between the zoomIn and zoomOut values i.e.
             * A x 1.25 = C but C * 0.8 no longer gives you A, and
             * vice versa.
            */
            if (pdf.bMap.Width > (pdf.panel.Width - 2))
            {
                if (zoomCnt < 0)
                {
                    pdf.hScrollPos = (int)((double)pdf.hScrollPos * Math.Pow(pdf.zoomInVal, (double)Math.Abs(zoomCnt)));
                }
                else
                {
                    pdf.hScrollPos = (int)((double)pdf.hScrollPos * Math.Pow(pdf.zoomOutVal, (double)zoomCnt));
                }

                pdf.panel.HorizontalScroll.Value = pdf.hScrollPos;
                pdf.panel.HorizontalScroll.Value = pdf.hScrollPos;
            }

            if (pdf.bMap.Height > (pdf.panel.Height - 2))
            {
                if (zoomCnt < 0)
                {
                    pdf.vScrollPos = (int)((double)pdf.vScrollPos * Math.Pow(pdf.zoomInVal, (double)Math.Abs(zoomCnt)));
                }
                else
                {
                    pdf.vScrollPos = (int)((double)pdf.vScrollPos * Math.Pow(pdf.zoomOutVal, (double)zoomCnt));
                }

                pdf.panel.VerticalScroll.Value = pdf.vScrollPos;
                pdf.panel.VerticalScroll.Value = pdf.vScrollPos;
            }

            DisableZoomResetBtns(ref pdf);
        }

        public void ProcessZoomResetBtns(ref Pdf pdf, ref PageArray pa)
        {
            if (pdf == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null ProcessPdf argument:\n\nProcessZoomResetBtns() received a null pdf argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            if (pa == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null PageArray argument:\n\nProcessZoomResetBtns() received a null pa argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            PgZoomData zd = null;

            pa.GetPgZoomDataRef((int)pdf.curPage - 1, ref zd);

            if (zd == null)
            {
                //using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, string.Format("ProcessZoomResetBtns():\n\npa.GetPgZoomDataRef({0}, ref zd) has returned a null PgZoomData.", pdf.curPage - 1)))
                //{
                //    mainMsg.ShowDialog(this);
                //}

                // Not necessarily an error here, we did not yet do a zoom in or out.
                return;
            }

            /*
             * It is possible to have a positive zd.zoomCnt and
             * have pdf.initZoomReset being true. This happens when
             * as an example, somebody does a Full Page or a Full Width
             * after they have done a Zoom In or a Zoom Out such that the
             * zd.zoomCnt is not zero.
             * 
             * The Full Width/Page operation will cause the 
             * pdf.initZoomReset to be set but they have not yet processed
             * it, which they do by choosing a Zoom In or a Zoom Out.
             * 
             * We would want that initialization to be done here.
             * There are other scenarios where this can happen.
             * 
            */
            if ((zd.zoomCnt == 0) || (pdf.initZoomReset))
            {
                DisableZoomResetBtns(ref pdf);
            }
            else
            {
                EnableZoomResetBtns(ref pdf);
            }
        }

        public void ProcessFirstPageNav(ref Pdf pdf, ref PageArray pa)
        {
            if (pdf == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null ProcessPdf argument:\n\nProcessFirstPageNav() received a null pdf argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            if (pa == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null PageArray argument:\n\nProcessFirstPageNav() received a null pa argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            if (pdf.pageNav == Pdf.pageNavMode.page)
            {
                pdf.curPage = 1;
            }
            else
            {
                // In boxed page navigation mode:
                pdf.boxedPageIndex = 0;
                pdf.curPage = pdf.boxedPageList[0];
            }

            pdf.ShowPage(ref docFile.hPdf, (int)pdf.curPage, true, ref pa);

            ProcessPageNum(ref pdf);

            ProcessZoomResetBtns(ref pdf, ref pa);
        }

        public void ProcessPrevPageNav(ref Pdf pdf, ref PageArray pa)
        {
            if (pdf == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null ProcessPdf argument:\n\nProcessPrevPageNav() received a null pdf argument."))
                {
                    mainMsg.ShowDialog(this);
                }
                return;
            }

            if (pa == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null PageArray argument:\n\nProcessPrevPageNav() received a null pa argument."))
                {
                    mainMsg.ShowDialog(this);
                }
                return;
            }

            if (pdf.pageNav == Pdf.pageNavMode.page)
            {
                pdf.curPage -= 1;
            }
            else
            {
                // In boxed page navigation mode:
                pdf.boxedPageIndex -= 1;
                pdf.curPage = pdf.boxedPageList[pdf.boxedPageIndex];
            }

            pdf.ShowPage(ref docFile.hPdf, (int)pdf.curPage, true, ref pa);

            ProcessPageNum(ref pdf);

            ProcessZoomResetBtns(ref pdf, ref pa);
        }

        public void ProcessNextPageNav(ref Pdf pdf, ref PageArray pa)
        {
            if (pdf == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null ProcessPdf argument:\n\nProcessNextPageNav() received a null pdf argument."))
                {
                    mainMsg.ShowDialog(this);
                }
                return;
            }

            if (pa == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null PageArray argument:\n\nProcessNextPageNav() received a null pa argument."))
                {
                    mainMsg.ShowDialog(this);
                }
                return;
            }

            if (pdf.pageNav == Pdf.pageNavMode.page)
            {
                pdf.curPage += 1;
            }
            else
            {
                // In boxed page navigation mode:
                pdf.boxedPageIndex += 1;
                pdf.curPage = pdf.boxedPageList[pdf.boxedPageIndex];
            }

            pdf.ShowPage(ref docFile.hPdf, (int)pdf.curPage, true, ref pa);

            ProcessPageNum(ref pdf);

            ProcessZoomResetBtns(ref pdf, ref pa);
        }

        public void ProcessLastPageNav(ref Pdf pdf, ref PageArray pa)
        {
            if (pdf == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null ProcessPdf argument:\n\nProcessLastPageNav() received a null pdf argument."))
                {
                    mainMsg.ShowDialog(this);
                }
                return;
            }

            if (pa == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null PageArray argument:\n\nProcessLastPageNav() received a null pa argument."))
                {
                    mainMsg.ShowDialog(this);
                }
                return;
            }

            if (pdf.pageNav == Pdf.pageNavMode.page)
            {
                pdf.curPage = pdf.numPages;
            }
            else
            {
                // In boxed page navigation mode:
                pdf.boxedPageIndex = pdf.boxedPageList.Count - 1;
                pdf.curPage = pdf.boxedPageList[pdf.boxedPageIndex];
            }

            pdf.ShowPage(ref docFile.hPdf, (int)pdf.curPage, true, ref pa);

            ProcessPageNum(ref pdf);

            ProcessZoomResetBtns(ref pdf, ref pa);
        }

        public void ProcessFullPage(ref Pdf pdf, ref PageArray pa)
        {
            if (pdf == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null ProcessPdf argument:\n\nProcessFullPage() received a null pdf argument."))
                {
                    mainMsg.ShowDialog(this);
                }
                return;
            }

            if (pa == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null PageArray argument:\n\nProcessFullPage() received a null pa argument."))
                {
                    mainMsg.ShowDialog(this);
                }
                return;
            }

            /*
             * This is a 'forced' dpi change, so we disable the
             * zoom reset capability.
            */

            pdf.ScaleRects(ref docFile.hPdf, "FullPage", ref pa);

            pdf.ShowPage(ref docFile.hPdf, (int)pdf.curPage, true, ref pa);

            DisableZoomResetBtns(ref pdf);
        }

        public void ProcessFullWidth(ref Pdf pdf, ref PageArray pa)
        {
            if (pdf == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null ProcessPdf argument:\n\nProcessFullWidth() received a null pdf argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            if (pa == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null PageArray argument:\n\nProcessFullWidth() received a null pa argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            /*
             * This is a 'forced' dpi change, so we disable the
             * zoom reset capability.
            */

            pdf.ScaleRects(ref docFile.hPdf, "FullWidth", ref pa);

            pdf.ShowPage(ref docFile.hPdf, (int)pdf.curPage, true, ref pa);

            DisableZoomResetBtns(ref pdf);
        }

        public void ProcessMinMaxRestore(ref Pdf pdf, ref PageArray pa)
        {
            if (pa == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null PageArray argument:\n\nProcessMinMaxRestore() received a null pa argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            /*
             * Caller has verified the pdf is not null.
             * 
             * This is called when we are minimizing, maximizing, or restoring
             * the form. We rescale the pages to make the most out of the available
             * panel area, and disable zoom reset capability.
            */

            /*
             * Restore visibility as we intercepted the call to maximize and turned this off.
             * This is the only way I could find that prevented the image from flashing
             * briefly in upper left corner of PicPanel.
            */
            if (this.WindowState == FormWindowState.Maximized)
            {
                pdf.panel.Visible = true;
            }

            if (pdf.panel.BackColor == SystemColors.Control)
            {
                if (pdf.ScaleRects(ref docFile.hPdf, "OpenFile", ref pa))
                {
                    if (pdf.numPages > 0)
                    {
                        pdf.ShowPage(ref docFile.hPdf, (int)pdf.curPage, true, ref pa);

                        DisableZoomResetBtns(ref pdf);
                    }
                }
            }
        }

        public void ProcessResize(ref Pdf pdf, ref PageArray pa)
        {
            if (pdf == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null ProcessPdf argument:\n\nProcessResize() received a null pdf argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            if (pa == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null PageArray argument:\n\nProcessResize() received a null pa argument."))
                {
                    mainMsg.ShowDialog(this);
                }

                return;
            }

            //MessageBox.Show(string.Format("{0}: ProcessResize()!", pdf.id));

            int scrollBarMask = 0;

            pdf.ScaleRects(ref docFile.hPdf, "Resize", ref pa);

            if (pdf.curPageZoomDelta)
            {
                /*
                 * The Resize ended up 'growing' the image for the current
                 * page which causes a forced dpi and bitmap settings change.
                 * 
                 * We Resize ALL pages like we do for all of the other
                 * ScaleRects() operations. We will disable zoom reset
                 * capability for the current page. Odds are this applies to
                 * all pages, but it is possible for a landscape page to
                 * not encounter a forced dpi change while the other pages
                 * do.
                */
                pdf.curPageZoomDelta = false;

                DisableZoomResetBtns(ref pdf);
            }

            if (this.WindowState == FormWindowState.Maximized)
            {
                /*
                 * Restore this as we intercepted the call to maximize and turned this off.
                 * This is the only way I could find that prevented the image from flashing
                 * briefly in upper left corner of PicPanel.
                */
                pdf.panel.Visible = true;
            }

            scrollBarMask = ((pdf.panel.ClientSize.Width > pdf.pictureBox.Width ? 10 : 0) + (pdf.panel.ClientSize.Height > pdf.pictureBox.Height ? 1 : 0));
            //this.Text = string.Format("{0} scrollBarMask: {1}", pdf.id, scrollBarMask);

            switch (scrollBarMask)
            {
                case (0):

                    if ((pdf.pictureBox.Location.X > 0) || (pdf.pictureBox.Location.Y > 0))
                    {
                        Point pt = pdf.pictureBox.Location;

                        //this.Text = string.Format("{0}: {1} | Pic.X > 0 {2} | Pic.Y > 0 {3}", pdf.id, scrollBarMask, (pdf.pictureBox.Location.X > 0), (pdf.pictureBox.Location.Y > 0));

                        if (pdf.pictureBox.Location.X > 0)
                        {
                            pt.X = 0;
                        }

                        if (pdf.pictureBox.Location.Y > 0)
                        {
                            pt.Y = 0;
                        }

                        pdf.pictureBox.Location = pt;
                    }
                    break;

                case (1):
                case (10):

                    pdf.resetScrollBars = false;
                    pdf.CenterPicPanel();
                    pdf.resetScrollBars = true;

                    break;

                case (11):

                    pdf.ShowPage(ref docFile.hPdf, (int)pdf.curPage, true, ref pa);

                    break;
            }
        }

        public void ProcessPageNavToolStripClick(Pdf.pageNavMode navMode, ref Pdf pdf, ref PageArray pa)
        {
            if (pdf == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null ProcessPdf argument:\n\nProcessPageNavToolStripClick() received a null pdf argument."))
                {
                    mainMsg.ShowDialog(this);
                }
                return;
            }

            if (pa == null)
            {
                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, "null PageArray argument:\n\nProcessPageNavToolStripClick() received a null pa argument."))
                {
                    mainMsg.ShowDialog(this);
                }
                return;
            }

            // Caller has ensured we are not already in this mode.
            pdf.pageNav = navMode;

            if (pdf.pageNav != Pdf.pageNavMode.page)
            {
                /*
                 * Switching to a boxed page navigation mode. Made sure these lists
                 * of page numbers by type have page numbers in them, and the various
                 * tool strip menu items for selecting a particular boxed page
                 * navigation have only been enabled if there are pages with those
                 * boxes on them.
                 */
                pdf.boxedPageList = pa.GetBoxedPageList(navMode.ToString());

                /*
                * We need to set the index based on the pdf.curPage value,
                * which can be any page, not necessarily a page with boxes.
                */
                if (pdf.curPage < pdf.boxedPageList[0])
                {
                    /*
                     * Default to the first boxed index position and move to that page.
                    */
                    pdf.boxedPageIndex = 0;
                    pdf.curPage = pdf.boxedPageList[0];
                    pdf.ShowPage(ref docFile.hPdf, (int)pdf.curPage, true, ref pa);
                }
                else if (pdf.curPage > pdf.boxedPageList[pdf.boxedPageList.Count - 1])
                {
                    /*
                     * Default to the last boxed index position and move to that page.
                    */
                    pdf.boxedPageIndex = pdf.boxedPageList.Count - 1;
                    pdf.curPage = pdf.boxedPageList[pdf.boxedPageIndex];
                    pdf.ShowPage(ref docFile.hPdf, (int)pdf.curPage, true, ref pa);
                }
                else
                {
                    /*
                     * Current page num is somewhere in the page range encompassed by the
                     * page numbers in the list. We need to determine if it is in the list,
                     * if so we use that index, we are already on that page. If not, we set
                     * for the nearest boxed index and move to that page.
                    */
                    if (pdf.boxedPageList.Contains(pdf.curPage))
                    {
                        pdf.boxedPageIndex = pdf.boxedPageList.IndexOf(pdf.curPage);
                    }
                    else
                    {
                        for (int i = 1; i < pdf.boxedPageList.Count; i++)
                        {
                            if (pdf.curPage < pdf.boxedPageList[i])
                            {
                                // We will autoselect the closest boxed page:
                                if ((pdf.boxedPageList[i] - pdf.curPage) < (pdf.curPage - pdf.boxedPageList[i - 1]))
                                {
                                    pdf.boxedPageIndex = i;
                                }
                                else
                                {
                                    pdf.boxedPageIndex = i - 1;
                                }

                                break;
                            }
                        }

                        pdf.curPage = pdf.boxedPageList[pdf.boxedPageIndex];
                        pdf.ShowPage(ref docFile.hPdf, (int)pdf.curPage, true, ref pa);
                    }
                }
            }

            ProcessPageNum(ref pdf);
            ProcessPageNavMode(pdf.pageNav, pdf.id);
        }

        static public Rectangle GetRectangle(Point p1, Point p2)
        {
            return new Rectangle(Math.Min(p1.X, p2.X), Math.Min(p1.Y, p2.Y),
                Math.Abs(p1.X - p2.X), Math.Abs(p1.Y - p2.Y));
        }

        public void ProcessSingleScrnCLick()
        {
            /*
             * User is switching from a split screen view to a single screen
             * view. We end up collapsing the pdf2 panel. The pdf1 panel always
             * exists in an uncollapsed state, it is the pdf2 panel that collapses
             * and uncollapses as the user clicks the screen split buttons.
            */
            int pdfMask;

            pdfMask = (pdf1 == null ? 0 : 10) + (pdf2 == null ? 0 : 1);

            splitScreen = false;

            splitContainer1.Panel2Collapsed = true;

            switch (pdfMask)
            {
                case (0):

                    // pdf1 and padf2 are both empty:
                    singleScrnBtn.Enabled = false;
                    singleScreenToolStripMenuItem.Enabled = false;

                    splitVertBtn.Enabled = false;
                    splitScreenVerticallyToolStripMenuItem.Enabled = false;

                    splitHorizBtn.Enabled = false;
                    splitScreenHorizontallyToolStripMenuItem.Enabled = false;

                    break;

                case (11):

                    // pdf1 is not empty, padf2 is not empty:
                    singleScrnBtn.Enabled = false;
                    singleScreenToolStripMenuItem.Enabled = false;

                    splitVertBtn.Enabled = true;
                    splitScreenVerticallyToolStripMenuItem.Enabled = true;

                    splitHorizBtn.Enabled = true;
                    splitScreenHorizontallyToolStripMenuItem.Enabled = true;

                    /*
                     * Going to rescale pdf1 to larger panel and will reset
                     * the 'zoom reset' buttons if enabled.
                    */
                    if (pdf1.ScaleRects(ref docFile.hPdf, "OpenFile", ref pa1))
                    {
                        if (pdf1.numPages > 0)
                        {
                            pdf1.ShowPage(ref docFile.hPdf, (int)pdf1.curPage, true, ref pa1);

                            DisableZoomResetBtns(ref pdf1);
                        }
                    }

                    pdf2.ClearBmaps();

                    DisableZoomResetBtns(ref pdf2);

                    pdf2 = null;
                    pa2 = null;

                break;
            }
        }

        public void ProcessSplitHorScrnCLick()
        {
            /*
             * If the screen is already split vertically: We will just do the horizontal
             * split with what the panels may or may not contain. If they are displaying
             * a pdf we will rescale each page's rectangles to account for the dimensional
             * change.
             * 
             * If the screen is not currently split: We will rescale the pdf1 rectangles
             * if the pdf1 is not null and also copy any pdf1 content to the pdf2 panel
             * rescaling each page's rectangles for that panel.
            */

            splitHorizBtn.Enabled = false;
            splitScreenHorizontallyToolStripMenuItem.Enabled = false;

            splitContainer1.Orientation = Orientation.Horizontal;
            splitContainer1.SplitterDistance = (int)(splitContainer1.ClientSize.Height * 0.5);

            if (splitScreen)
            {
                splitVertBtn.Enabled = true;
                splitScreenVerticallyToolStripMenuItem.Enabled = true;

                if (pdf2 != null)
                {
                    if (pa2 != null)
                    {
                        // Rescale all pdf2 page boxes to reflect new panel dimension/size.
                        if (pdf2.ScaleRects(ref docFile.hPdf, "OpenFile", ref pa2))
                        {
                            if (pdf2.numPages > 0)
                            {
                                pdf2.ShowPage(ref docFile.hPdf, (int)pdf2.curPage, true, ref pa2);

                                DisableZoomResetBtns(ref pdf2);
                            }
                        }
                    }
                }
            }
            else
            {
                splitScreen = true;

                /*
                 * Form view will be split horizontally. We will set the 2
                 * display panels to each consume 50% of the viewable Form
                 * area. Each pdf will be re-scaled to fit the new reduced
                 * area if a pdf is being displayed.
                */
                singleScrnBtn.Enabled = true;
                singleScreenToolStripMenuItem.Enabled = true;

                splitContainer1.Panel2Collapsed = false;

                if (pdf1 != null)
                {
                    /*
                     * Use pdf1 to seed pdf2. true means we also want to set
                     * .isSelected, .isCalculated, and .isCalculating fields.
                    */
                    SetupPdf2(docFile.zipFile, true);
                }
            }

            if (pdf1 != null)
            {
                /*
                 * Rescale all page boxes to reflect new panel dimension/size.
                 * 
                 * ScaleRects(pdfFile, "OpenFile", ref pa1) signals a reset of the PgZoomData
                 * for each page since we are moving to a new panel size, so for now we
                 * will disable the zoom reset buttons. That signalling is caught by
                 * the zoom in and zoom out routines.
                */
                if (pa1 != null)
                {
                    if (pdf1.ScaleRects(ref docFile.hPdf, "OpenFile", ref pa1))
                    {
                        if (pdf1.numPages > 0)
                        {
                            pdf1.ShowPage(ref docFile.hPdf, (int)pdf1.curPage, true, ref pa1);

                            DisableZoomResetBtns(ref pdf1);
                        }
                    }
                }
            }
        }

        public void ProcessSplitVerScrnCLick()
        {
            /*
             * If the screen is already split horizontally: We will just do the vertical
             * split with what the panels may or may not contain. If they are displaying
             * a pdf we will rescale each page's rectangles to account for the dimensional
             * change.
             * 
             * If the screen is not currently split: We will rescale the pdf1 rectangles
             * if the pdf1 is not null and also copy any pdf1 content to the pdf2 panel
             * rescaling each page's rectangles for that panel.
            */

            splitVertBtn.Enabled = false;
            splitScreenVerticallyToolStripMenuItem.Enabled = false;

            splitContainer1.Orientation = Orientation.Vertical;
            splitContainer1.SplitterDistance = (int)(splitContainer1.ClientSize.Width * 0.5);

            if (splitScreen)
            {
                splitHorizBtn.Enabled = true;
                splitScreenHorizontallyToolStripMenuItem.Enabled = true;

                if (pdf2 != null)
                {
                    if (pa2 != null)
                    {
                        // Rescale all pdf2 page boxes to reflect new panel dimension/size.
                        if (pdf2.ScaleRects(ref docFile.hPdf, "OpenFile", ref pa2))
                        {
                            if (pdf2.numPages > 0)
                            {
                                pdf2.ShowPage(ref docFile.hPdf, (int)pdf2.curPage, true, ref pa2);

                                DisableZoomResetBtns(ref pdf2);
                            }
                        }
                    }
                }
            }
            else
            {
                splitScreen = true;

                /*
                 * Form view will be split vertically. We will set the 2
                 * display panels to each consume 50% of the viewable Form
                 * area. Each pdf will be re-scaled to fit the new reduced
                 * area if a pdf is being displayed.
                */
                singleScrnBtn.Enabled = true;
                singleScreenToolStripMenuItem.Enabled = true;

                splitContainer1.Panel2Collapsed = false;

                if (pdf1 != null)
                {
                    /*
                     * Use pdf1 to seed pdf2. true means we also want to set
                     * .isSelected, .isCalculated, and .isCalculating fields.
                    */
                    SetupPdf2(docFile.zipFile, true);
                }
            }

            if (pdf1 != null)
            {
                /*
                 * Rescale all page boxes to reflect new panel dimension/size.
                 * 
                 * ScaleRects(pdfFile, "OpenFile", ref pa1) signals a reset of the PgZoomData
                 * for each page since we are moving to a new panel size, so for now we
                 * will disable the zoom reset buttons. That signalling is caught by
                 * the zoom in and zoom out routines.
                */
                if (pa1 != null)
                {
                    if (pdf1.ScaleRects(ref docFile.hPdf, "OpenFile", ref pa1))
                    {
                        if (pdf1.numPages > 0)
                        {
                            pdf1.ShowPage(ref docFile.hPdf, (int)pdf1.curPage, true, ref pa1);

                            DisableZoomResetBtns(ref pdf1);
                        }
                    }
                }
            }
        }

        /*
         * This is used in the lasso selection - left mouse.
         * 
         * Called By:
         * 
         *  pictureBox1_MouseDown()
         *  pictureBox2_MouseDown()
         */
        public void ProcessLeftMouseDown(ref Pdf pdf,  Point location)
        {
            /*
             * Caller has ensured that pdf is not null and it is a left
             * mouse button event.
            */
            pdf.mDown = location;
            pdf.mLassoDown = true;
        }

        /*
         * This is used in the lasso selection - left mouse.
         * 
         * Called By:
         * 
         *  pictureBox1_MouseMove()
         *  pictureBox2_MouseMove()
         *  
         * We are allowing the selection rectangle to be drawn right now
         * because we are not sure what the user is selecting/unselecting
         * and there are selection 'rules' for things we can not discern
         * right now. We do not want to popup error windows at this point.
         */
        public void ProcessLeftMouseMove(ref Pdf pdf, Point location)
        {
            /*
             * Caller has ensured that pdf is not null and it is a left
             * mouse button event.
            */
            if (pdf.mLassoDown)
            {
                pdf.mLassoMove = true;

                using (Graphics g = pdf.pictureBox.CreateGraphics())
                {
                    Pen penSel = new Pen(Color.Black, 2f);
                    penSel.DashStyle = System.Drawing.Drawing2D.DashStyle.Dot;

                    pdf.pictureBox.Refresh();
                    pdf.lassoRect = GetRectangle(pdf.mDown, location);
                    g.DrawRectangle(penSel, pdf.lassoRect);
                }
            }
        }

        /*
         * User has clicked with the left mouse button. We determine if
         * they clicked a box and if so, are they selecting or unselecting
         * the box? If they are selecting the box we will send it on it's
         * way to the Vscode editor either directly or vis the Tool Belt's
         * Box Selector.
         * 
         * Called By:
         * 
         *  pictureBox1_MouseClick()
         *  pictureBox2_MouseClick()
         *  
         * If a rightCtrlBoxSel box was clicked, we will reset this
         * selection and set it for a left mouse selection.
         * 
         * If a rightCalcBoxSel box was clicked:
         * 
         * We will keep this selection status and also set it as a
         * left mouse selection, they both share the same selection
         * rectangle style. Doing this so this box can be sent to
         * Vscode editor for it's own calc, not too common, but there
         * are cases where a box being calculated is also on the right
         * side of the equals sign.
         * 
         * The rightCalcBoxSel flag for the box makes sure it is
         * colored to indicate a box that is being calculated.
         */
        public void ProcessLeftMouseClick(ref Pdf pdf, ref PageArray pa, Point location)
        {
            /*
             * Caller has ensured that pdf, pa and ctClient are not null,
             * and pdf.mLassoMove is false, and it is a left mouse button event.
            */
            pdf.mLassoDown = false;

            List<Field> fa = null;
            List<LeftBoxSel> boxSelList = new List<LeftBoxSel>();
            LeftBoxSel lbs;
            bool selecting = true;
            bool chkBoxCtrlSel = false;
            bool dfRetSel = false;
            bool docSel = false;
            bool error = false;
            int paIndex = (int)(pdf.curPage - 1);
            string msg = "";

            pa.GetFieldArrayRef(paIndex, ref fa);

            if (fa != null)
            {
                for (int i = 0; i < fa.Count; i++)
                {
                    if (fa[i].sbox.Contains(location))
                    {
                        // Set whether we are selecting or unselecting this box.
                        selecting = !fa[i].leftBoxSel;

                        /*
                         * If selecting, check if we can select boxes:
                         */
                        if (selecting == true)
                        {
                            chkBoxCtrlSel = (ctClient.boxSelClient == boxSelDest.chkBoxCtrl);
                            dfRetSel = (ctClient.boxSelClient == boxSelDest.dfRetriever);
                            docSel = (ctClient.boxSelClient == boxSelDest.document);

                            if (!chkBoxCtrlSel && (ctClient.vscodeLock || (!dfRetSel && !docSel)))
                            {
                                msg = "Box Selection Error:\n\nCalc Helper's Box Selection Mode is OFF (red 'X').\n\nBox will not be selected.";
                                using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, msg))
                                {
                                    mainMsg.ShowDialog(this);
                                }

                                error = true;
                                break;
                            }

                            /*
                             * Now 'get the details' and store in our local list but we
                             * also need to take into account where this selection is
                             * headed - to Vscode, to a calc tool like Checkbox Control
                             * for example.
                             * 
                             * If unselecting we will store the box name and have
                             * ResetSelections() handle it.
                             */
                            lbs = new LeftBoxSel(fa[i].ref_name, paIndex, i);

                            if (chkBoxCtrlSel == true)
                            {
                                // Only take a non-controlled checkbox.
                                Field field = fa[i];

                                if (field is TextField)
                                {
                                    if (((TextField)field).checkbox == true)
                                    {
                                        if (field.ctrlType == -1)
                                        {
                                            boxSelList.Add(lbs);
                                        }
                                        else
                                        {
                                            /*
                                             * Tell user we do not Checkbox Control checkboxes
                                             * that are already controlled.
                                             */
                                            msg = string.Format("Bad Checkbox Selection:\n\nYou cannot select a controlled checkbox.\n\n{0} will be ignored.", field.ref_name);

                                            using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, msg))
                                            {
                                                mainMsg.ShowDialog(this);
                                            }

                                            break;
                                        }
                                    }
                                }
                            }
                            else
                            {
                                if (fa[i].rightCtrlBoxSel == true)
                                {
                                    fa[i].rightCtrlBoxSel = false;
                                    boxSel.rightCtrlBoxSel = null;
                                }

                                boxSelList.Add(lbs);
                            }
                        }
                        else
                        {
                            boxSel.unselLeftBoxes.Add(fa[i].ref_name);
                        }

                        break;
                    }
                }

                if (error == false)
                {
                    if (selecting == true)
                    {
                        /*
                         * Safety check..
                         */
                        if (boxSelList.Count > 0)
                        {
                            if (boxSel.ProcessLeftBoxSel(ref boxSelList, ref pdf1, ref pa1, ref pdf2, ref pa2) == true)
                            {
                                RefreshPanels();
                                SendOffLeftBoxSel(ref boxSelList);
                            }
                        }
                    }
                    else
                    {
                        // Probably overboard here but....
                        if (Pa1OrPdf1AreNull("ProcessLeftMouseClick", "Can not handle box unselection.") == false)
                        {
                            boxSel.ResetSelections(boxSelTypes.leftSel, ref pa1, ref pdf1, ref pa2, ref pdf2);
                            RefreshPanels();
                        }
                    }
                }
            }
        }

        /*
         * This is used in the lasso selection - left mouse.
         * 
         * Called By:
         * 
         *  pictureBox1_MouseUp()
         *  pictureBox2_MouseUp()
         * 
         * Now we can figure out what was lassoed. We only allow
         * either selections or unselections, not a mix of the two.
         * 
         * If a rightCtrlBoxSel box was lassoed, we will reset this
         * selection and set it for a left mouse selection.
         * 
         * If a rightCalcBoxSel box was lassoed:
         * 
         * We will keep this selection status and also set it as a
         * left mouse selection, they both share the same selection
         * rectangle style. Doing this so this box can be sent to
         * Vscode editor for it's own calc, not too common, but there
         * are cases where a box being calculated is also on the right
         * side of the equals sign.
         * 
         * The rightCalcBoxSel flag for the box makes sure it is
         * colored to indicate a box that is being calculated.
         */
        public void ProcessLeftMouseUp(ref Pdf pdf, ref PageArray pa)
        {
            /*
             * Caller has ensured that pdf, pa, and ctClient are not null,
             * and it is a left mouse button event.
            */
            if (pdf.mLassoMove)
            {
                List<Field> fa = null;
                List<LeftBoxSel> boxSelList = new List<LeftBoxSel>();
                LeftBoxSel lbs;
                bool selecting = true;
                bool chkBoxCtrlSel = false;
                bool dfRetSel = false;
                bool docSel = false;
                int cnt = 0;
                int paIndex = (int)(pdf.curPage - 1);
                string msg = "";

                /*
                 * Need to set these here [actually just chkBoxCtrlSel but might
                 * as well do all 3] even though they are needed for selecting
                 * boxes - we need to communicate with the Calc Tool Server so we
                 * can send the boxes to their destination.
                 * 
                 * When we switch to one of these box selection modes we first
                 * unselect all selections on the document.
                 */
                chkBoxCtrlSel = (ctClient.boxSelClient == boxSelDest.chkBoxCtrl);
                dfRetSel = (ctClient.boxSelClient == boxSelDest.dfRetriever);
                docSel = (ctClient.boxSelClient == boxSelDest.document);

                pa.GetFieldArrayRef(paIndex, ref fa);

                if (fa != null)
                {
                    for (int i = 0; i < fa.Count; i++)
                    {
                        if (fa[i].sbox.IntersectsWith(pdf.lassoRect))
                        {
                            if (chkBoxCtrlSel == true)
                            {
                                /*
                                 * We only want to count intersections with
                                 * non-controlled checkboxes.
                                 */
                                Field field = fa[i];

                                if (field is TextField)
                                {
                                    if (((TextField)field).checkbox == true)
                                    {
                                        if (field.ctrlType == -1)
                                        {
                                            cnt += 1;
                                        }
                                        else
                                        {
                                            /*
                                             * Tell user we do not Checkbox Control checkboxes
                                             * that are already controlled.
                                             */
                                            msg = string.Format("Bad Checkbox Lasso:\n\nYou cannot lasso a controlled checkbox.\n\n{0} will be ignored.", field.ref_name);

                                            using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, msg))
                                            {
                                                mainMsg.ShowDialog(this);
                                            }

                                            continue;
                                        }
                                    }
                                    else
                                    {
                                        continue;
                                    }
                                }
                                else
                                {
                                    continue;
                                }
                            }
                            else
                            {
                                // Count any box intersection.
                                cnt += 1;
                            }

                            if (cnt > 1)
                            {
                                // Make sure we do not mix and match regarding selecting vs. unselecting.
                                if ((selecting && fa[i].leftBoxSel) || (!selecting && !fa[i].leftBoxSel))
                                {
                                    msg = string.Format("Bad {0} Lasso:\n\nAll boxes must be either selected or unselected.\n\nAborting the operation.", (chkBoxCtrlSel == true ? "Checkbox" : ""));

                                    using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, msg))
                                    {
                                        mainMsg.ShowDialog(this);
                                    }

                                    cnt = 0;
                                    break;
                                }
                            }
                            else
                            {
                                // First intersection - set whether we are selecting or unselecting boxes.
                                selecting = !fa[i].leftBoxSel;

                                /*
                                 * If selecting, check if we can select boxes:
                                 */
                                if (selecting == true)
                                {
                                    if (!chkBoxCtrlSel && (ctClient.vscodeLock || (!dfRetSel && !docSel)))
                                    {
                                        msg = "Box Selection Error:\n\nCalc Helper's Box Selection Mode is OFF (red 'X').\n\nBoxes will not be selected.";
                                        using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, msg))
                                        {
                                            mainMsg.ShowDialog(this);
                                        }

                                        cnt = 0;
                                        break;
                                    }
                                }
                            }

                            /*
                             * If selecting we will 'get the details' and store in our local list
                             * but we also need to take into account where these selections are
                             * headed - to Vscode, to a calc tool like Checkbox Control for example.
                             * 
                             * If unselecting we will store the box name(s) and have ResetSelections()
                             * handle them.
                             */
                            if (selecting == true)
                            {
                                lbs = new LeftBoxSel(fa[i].ref_name, paIndex, i);

                                /*
                                 * We don't need to worry about if we are in Checkbox
                                 * Control box selection mode here as we skip over any
                                 * controlled checkboxes when in that mode.
                                 */
                                if (fa[i].rightCtrlBoxSel == true)
                                {
                                    fa[i].rightCtrlBoxSel = false;
                                    boxSel.rightCtrlBoxSel = null;
                                }

                                boxSelList.Add(lbs);
                            }
                            else
                            {
                                boxSel.unselLeftBoxes.Add(fa[i].ref_name);
                            }
                        }
                    }

                    if (cnt > 0)
                    {
                        if (selecting == true)
                        {
                            /*
                             * Just a safety check here, should be OK but you never know.
                             */
                            if (boxSelList.Count > 0)
                            {
                                if (boxSel.ProcessLeftBoxSel(ref boxSelList, ref pdf1, ref pa1, ref pdf2, ref pa2) == true)
                                {
                                    RefreshPanels();
                                    SendOffLeftBoxSel(ref boxSelList);
                                }
                                else
                                {
                                    cnt = 0;
                                }
                            }
                            else
                            {
                                cnt = 0;
                            }
                        }
                        else
                        {
                            // Probably overboard here but....
                            if (Pa1OrPdf1AreNull("ProcessLeftMouseUp", "Can not handle box unselection(s).") == true)
                            {
                                cnt = 0;
                            }
                            else
                            {
                                boxSel.ResetSelections(boxSelTypes.leftSel, ref pa1, ref pdf1, ref pa2, ref pdf2);
                                RefreshPanels();
                            }
                        }
                    }
                }

                if (cnt == 0)
                {
                    // Clear any pending unselects since we are not doing it.
                    if (boxSel.unselLeftBoxes.Count > 0)
                    {
                        boxSel.unselLeftBoxes.Clear();
                    }

                    // Killing the rect like this rather than via pdf.ShowPage().
                    pdf.pictureBox.Refresh();
                }

                pdf.mLassoMove = false;
                pdf.mLassoDown = false;
            }
        }

        /*
         * Boxes have been selected via either a left mouse button
         * click or lasso. This routine takes these boxes and moves
         * them along towards their final destination, which may be:
         * 
         * 1) The Tool Belt's Box Selector.
         * 2) To Vscode via a ctClient message.
         * 3) To another calc tool like Checkbox Control or
         *      Dataflow Retriever via a ctClient message.
         * 
         * Called By:
         * 
         *  ProcessLeftMouseUp()
         *  ProcessLeftMouseClick()
         * 
         * Caller has verified that boxSelList has at least one item
         * and ctClient is not null.
         */
        public void SendOffLeftBoxSel(ref List<LeftBoxSel> boxSelList)
        {
            bool selDone = false;
            string boxStr = "";
            string opStr = "";
            string formatStr = "";
            BoxSelArgs bsa;

            // First determine where these boxes are headed:
            switch (ctClient.boxSelClient)
            {
                case (boxSelDest.chkBoxCtrl):
                case (boxSelDest.dfRetriever):

                    foreach (LeftBoxSel lbs in boxSelList)
                    {
                        boxStr += lbs.box + ',';
                    }

                    boxStr.TrimEnd(',');
                    ctClient.SendBoxSelections(boxStr);
                    selDone = true;

                    break;
            }

            if (selDone == true)
            {
                return;
            }

            /*
             * We set this up in the calling function such that
             * being here [meaning we did not return just above]
             * means we are sending boxes to the document in the
             * Vscode editor.
             */
            if (boxSelList.Count > 1)
            {
                if (boxSel.bsOpLockState == boxSelOpLock.none)
                {
                    /*
                     * No operator lock, so we need the Tool Belt's
                     * Box Selector for this.
                     */
                    if (toolBelt == null)
                    {
                        ToolBeltStartUp(ToolBelt.tbTabPage.boxSel);
                    }
                    //else if (toolBelt.tabPage != ToolBelt.tbTabPage.boxSel)
                    //{
                    //    toolBelt.SelectTabPage(ToolBelt.tbTabPage.boxSel);
                    //}

                    
                    // Trying to 'bring up' the toolBelt if minimized - START:
                    else
                    {
                        if (toolBelt.tabPage != ToolBelt.tbTabPage.boxSel)
                        {
                            toolBelt.SelectTabPage(ToolBelt.tbTabPage.boxSel);
                        }

                        if (toolBelt.WindowState == FormWindowState.Minimized)
                        {
                            toolBelt.WindowState = FormWindowState.Normal;
                        }

                        toolBelt.Activate();
                    }
                    // Trying to 'bring up' the toolBelt if minimized - STOP:




                    bsa = new BoxSelArgs();
                    bsa.cmd = "MultiBoxWait";
                    bsa.boxes = new List<string>();

                    foreach (LeftBoxSel lbs in boxSelList)
                    {
                        bsa.boxes.Add(lbs.box);
                    }

                    OnBoxSel(bsa);

                    selDone = true;
                }
                else
                {
                    // Tool Belt Box Selector has an operator lock:

                    if (toolBelt != null)
                    {
                        /*
                         * Tool Belt Box Selector has an operator lock
                         * but we need to check if a one-shot button is
                         * setup for the next box selection(s) which is
                         * this one.
                         * 
                         * If so let the Tool Belt's Box Selector handle
                         * this.
                         */
                        if (toolBelt.bsOneShotState != ToolBelt.boxSelOneShot.none)
                        {
                            if (toolBelt.tabPage != ToolBelt.tbTabPage.boxSel)
                            {
                                toolBelt.SelectTabPage(ToolBelt.tbTabPage.boxSel);
                            }

                            bsa = new BoxSelArgs();
                            bsa.cmd = "MultiBoxOpLockOneShot";
                            bsa.boxes = new List<string>();

                            foreach (LeftBoxSel lbs in boxSelList)
                            {
                                bsa.boxes.Add(lbs.box);
                            }

                            OnBoxSel(bsa);

                            selDone = true;
                        }
                    }

                    if (selDone == false)
                    {
                        /*
                         * We will compose the multi box selection string using
                         * the locked operator and the Tool Belt's Box
                         * Selector's current box format and send it on
                         * to Vscode via ctClient.
                         * 
                         * First set the operator string:
                         */
                        switch (boxSel.bsOpLockState)
                        {
                            case (boxSelOpLock.add):
                                opStr = " + ";
                                break;

                            case (boxSelOpLock.div):
                                opStr = " / ";
                                break;

                            case (boxSelOpLock.mult):
                                opStr = " * ";
                                break;

                            case (boxSelOpLock.sub):
                                opStr = " - ";
                                break;
                        }

                        formatStr = boxSel.boxSelFormats[boxSel.boxSelFormatsIndex];

                        for (int i = 0; i < boxSelList.Count; i++)
                        {
                            if (i < boxSelList.Count - 1)
                            {
                                boxStr += (string.Format(formatStr, boxSelList[i].box) + opStr);
                            }
                            else
                            {
                                boxStr += string.Format(formatStr, boxSelList[i].box);
                            }
                        }

                        ctClient.SendBoxSelections(boxStr);
                        selDone = true;
                    }
                }
            }
            else
            {
                /* 
                 * Single box selection. Tool Belt Box Selector's
                 * Operator Lock does not apply for a single box selection.
                 */
                if (toolBelt != null)
                {
                    /* boxSel.toolBeltOpWaitBoxes.Count > 0:
                     * 
                     * Tool Belt's Box Selector is waiting for the user to
                     * click a separator or operator button to use with these
                     * boxes. The one shot button status does not matter here
                     * since we are not sending these boxes on to Vscode yet.
                     * 
                     * toolBelt.bsOneShotState != ToolBelt.boxSelOneShot.none:
                     * 
                     * Tool Belt's Box Selector has a one shot button set,
                     * so send this box on to it so it can handle it.
                     */
                    if ((boxSel.toolBeltOpWaitBoxes.Count > 0) || (toolBelt.bsOneShotState != ToolBelt.boxSelOneShot.none))
                    {
                        if (toolBelt.tabPage != ToolBelt.tbTabPage.boxSel)
                        {
                            toolBelt.SelectTabPage(ToolBelt.tbTabPage.boxSel);
                        }

                        bsa = new BoxSelArgs();
                        bsa.cmd = (boxSel.toolBeltOpWaitBoxes.Count > 0 ? "SingleBoxWait" : "SingleBoxOneShot");
                        bsa.boxes = new List<string>();
                        bsa.boxes.Add(boxSelList[0].box);

                        OnBoxSel(bsa);

                        selDone = true;
                    }
                }

                if (selDone == false)
                {
                    /*
                     * Prepare boxStr using the current box format and send
                     * on to Vscode via ctClient.
                     */
                    boxStr = string.Format(boxSel.boxSelFormats[boxSel.boxSelFormatsIndex], boxSelList[0].box);
                    ctClient.SendBoxSelections(boxStr);
                }
            }
        }

        /*
         * This starts up the Tool Belt and selects the tab page.
         * 
         * Called From:
         * 
         * SendOffLeftBoxSel()
         */
        public void ToolBeltStartUp(ToolBelt.tbTabPage tabPage)
        {
            toolBelt = new ToolBelt(ref boxSel, ref ctClient, ref docFile, ref pa1, tabPage);
            docFile.tb = toolBelt;

            /*
             * Subscribe to the ToolBelt's event so we can receive commands
             * or info from it, like unselecting boxes.
             */
            ToolBelt.ToolBeltEvent += toolBelt_ToolBeltEvent;

            toolBelt.Show(this);
            toolBelt.FormClosed += new FormClosedEventHandler(ToolBeltClosed);
        }
    }
}